# Jupyter Notebook Bevezető
A Jupyter Notebook interaktív felületet nyújt a kódoláshoz. A _JOKER_ szerveren jelenleg csak a **Python 3**-as kernel érhető el, de nagyon sok más nyelvhez is használható. 

Egy-egy ilyen notebook tartalmazhat futtatandó kódot és annak kimenetét, valamint a kód leírását, megértését megkönnyítő markdown kiterjesztést. A notebook fájl html/pdf/latex... dokumentumba is kimenthető, így pl. egy beadandónál, nagyon könnyen felhasználható. Ez a doksi is ott készült. 

A _JOKER_ szerver érdekessége, hogy a Jupyter szervert nem a helyi gépen futtatjuk, hanem a _JOKER_ -en, így elérjük a _JOKER_ által biztosított erőforrásokat, mint pl. GPU számítási kapacitást is, bárhonnan. 
## Kezelés
A kezelést nagyon nem is ragoznám, a Help menü alatt megtekinthetőek a fontosabb gyorsbillentyűk, illetve egy kezdetleges tutorial-t is indíthatunk.
## Hello Jupyter Notebook
Használatkor a hagyományos python-os interpreter mintájára működik, azaz tetszőlegesen tudunk a **Code** szekciókba python-os vagy más kódot elhelyezni és azt lefuttatni, pl.:


```python
print('Hello Jupyter Notebook')
```


```python
a = 10
```


```python
print('a: ', a)
```

## Magic
Van lehetőségünk további beépített parancsok használatára, melyeket most nem részletezek, néhány példát szeretnék mutatni: 


```python
lsmagic
```


```python
%%HTML 
<a href="https://www.vik.bme.hu/">BME VIK Honlap</a>
```


```latex
%%latex
$\sum_{n=1}^{\infty} \frac{1}{2^{n}} = 1$
```

## Mandelbrot halmaz
Komplex számokból álló sík, melyre:
$$x_1 = 0$$
$$x_{n+1}=x_n^2+c$$

Forrás: [link](https://levelup.gitconnected.com/mandelbrot-set-with-python-983e9fc47f56)


```python
# Kiszámolós függvény:
import matplotlib.pyplot as plt
import numpy as np
def get_iter(c:complex, thresh:int =4, max_steps:int =25) -> int:
    z=c
    i=1
    while i<max_steps and (z*z.conjugate()).real<thresh:
        z=z*z +c
        i+=1
    return i
```


```python
# Kirajzoló függvény
def plotter(n, thresh, max_steps=25):
    mx = 2.48 / (n-1)
    my = 2.26 / (n-1)
    mapper = lambda x,y: (mx*x - 2, my*y - 1.13)
    img=np.full((n,n), 255)
    for x in range(n):
        for y in range(n):
            it = get_iter(complex(*mapper(x,y)), thresh=thresh, max_steps=max_steps)
            img[y][x] = 255 - it
    return img
```


```python
n=3000
img = plotter(n, thresh=4, max_steps=50)
plt.imshow(img, cmap="plasma")
plt.axis("off")
plt.show()
```

## GPU támogatott Mandelbrot
Helyezzük el a @jit annotációt a számítási függvényekhez. Meg kell jegyezni, hogy az annotáció a pythonos kódot natívra is cseréli, nemcsak a GPU-t használja.


```python
# Kiszámolós függvény:
import matplotlib.pyplot as plt
import numpy as np
from numba import jit
@jit
def get_iter(c:complex, thresh:int =4, max_steps:int =25) -> int:
    z=c
    i=1
    while i<max_steps and (z*z.conjugate()).real<thresh:
        z=z*z +c
        i+=1
    return i
```


```python
# Kirajzoló függvény
@jit
def plotter(n, thresh, max_steps=25):
    mx = 2.48 / (n-1)
    my = 2.26 / (n-1)
    mapper = lambda x,y: (mx*x - 2, my*y - 1.13)
    img=np.full((n,n), 255)
    for x in range(n):
        for y in range(n):
            it = get_iter(complex(*mapper(x,y)), thresh=thresh, max_steps=max_steps)
            img[y][x] = 255 - it
    return img
```


```python
n=3000
img = plotter(n, thresh=4, max_steps=50)
plt.imshow(img, cmap="plasma")
plt.axis("off")
plt.show()
# sokkal gyorsabban kapunk eredményt :)
```


```python
from numba import jit
import numpy as np  
from timeit import default_timer as timer    

def func(a):                                 
    for i in range(10000000): 
        a[i]+= 1      

 
@jit                          
def func2(a): 
    for i in range(10000000): 
        a[i]+= 1
if __name__=="__main__": 
    n = 10000000                            
    a = np.ones(n, dtype = np.float64) 
    b = np.ones(n, dtype = np.float32) 

    start = timer() 
    func(a) 
    print("without GPU:", timer()-start)     

    start = timer() 
    func2(a) 
    print("with GPU:", timer()-start) 
```

## PyTorch
[link](https://pytorch.org/tutorials/beginner/basics/quickstart_tutorial.html)



```python
import torch
from torch import nn
from torch.utils.data import DataLoader
from torchvision import datasets
from torchvision.transforms import ToTensor, Lambda, Compose
import matplotlib.pyplot as plt
```


```python
# Download training data from open datasets.
training_data = datasets.FashionMNIST(
    root="data",
    train=True,
    download=True,
    transform=ToTensor(),
)
# Download test data from open datasets.
test_data = datasets.FashionMNIST(
    root="data",
    train=False,
    download=True,
    transform=ToTensor(),
)
```


```python
batch_size = 64

# Create data loaders.
train_dataloader = DataLoader(training_data, batch_size=batch_size)
test_dataloader = DataLoader(test_data, batch_size=batch_size)

for X, y in test_dataloader:
    print("Shape of X [N, C, H, W]: ", X.shape)
    print("Shape of y: ", y.shape, y.dtype)
    break
```

    Shape of X [N, C, H, W]:  torch.Size([64, 1, 28, 28])
    Shape of y:  torch.Size([64]) torch.int64


_HF befejezni_

## Osztályozás TensorFlow-val
Telepítsük a TensorFlow-t, egyszer majd talán központilag le lesz töltve, így nem kell magunknak leszedni:
```
module load anaconda3
conda create -n tf-gpu tensorflow-gpu
conda activate tf-gpu
```
Nézzük meg ezt a csomagot, és kövessük végig a leckét a JOKER szerveren: [image classification](https://www.tensorflow.org/tutorials/images/classification)


```python
#importok behúzása
import matplotlib.pyplot as plt
import numpy as np
import os
import PIL
import tensorflow as tf

from tensorflow import keras
from tensorflow.keras import layers
from tensorflow.keras.models import Sequential
```
